#include <rapidcheck/catch.h>
#ifdef EXAMPLE_SOLUTION
#include "link_n_list_solution.h"
#else
#include "link_n_list.h"
#endif

#include <string>

// test with smaller blocs so list does not need to be so long
template< typename T >
using test_list = link_n_list< T, 4 >;

struct NotConstructible {
    NotConstructible() = delete;
};

TEST_CASE( "ctor" ) {
    test_list< int > ll;
    test_list< int > ll2( ll ); // NOLINT
    test_list< NotConstructible > lla;
}

TEST_CASE( "singleton" ) {
    test_list< int > ll;
    REQUIRE( ll.empty() );
    REQUIRE( ll.push_back( 1 ) == 1 );

    REQUIRE( !ll.empty() );
    REQUIRE( ll.front() == 1 );
    REQUIRE( ll.back() == 1 );
    REQUIRE( &ll.front() == &ll.back() );

    test_list< int > ll2( ll );
    REQUIRE( !ll2.empty() );
    REQUIRE( ll2.front() == 1 );
    REQUIRE( ll2.back() == 1 );
    REQUIRE( &ll2.front() == &ll2.back() );
    REQUIRE( &ll2.front() != &ll.front() );
    REQUIRE( &ll2.back() != &ll.back() );
}

TEST_CASE( "nontrivial" ) {
    // this is primarily for leak detection
    test_list< std::unique_ptr< int > > list;
    list.push_back( std::make_unique< int >( 1 ) );
    list.push_back( std::make_unique< int >( 2 ) );
    REQUIRE( **list.begin() == 1 );
    REQUIRE( **std::next( list.begin() ) == 2 );
}

TEST_CASE( "size2" ) {
    test_list< int > ll;
    REQUIRE( ll.push_back( 1 ) == 1 );
    REQUIRE( ll.push_back( 2 ) == 2 );

    REQUIRE( !ll.empty() );
    REQUIRE( ll.front() == 1 );
    REQUIRE( ll.back() == 2 );
    REQUIRE( &ll.front() != &ll.back() );
}

TEST_CASE( "random" ) {
    // this test automatically generates vectors of integers, inserts them to
    // our list and compares results
    rc::prop( "random", []( std::vector< unsigned > to_insert ) {
        test_list< unsigned > list;
        std::copy( to_insert.begin(), to_insert.end(), std::back_inserter( list ) );

        RC_ASSERT( to_insert.empty() == list.empty() );

        if ( to_insert.empty() )
            return;

        RC_ASSERT( to_insert.front() == list.front() );
        RC_ASSERT( to_insert.back() == list.back() );

        test_list< unsigned > copy( list ); // NOLINT

        RC_ASSERT( to_insert.front() == copy.front() );
        RC_ASSERT( to_insert.back() == copy.back() );

        auto check = to_insert.begin();
        for ( auto &v : list ) {
            RC_ASSERT( v == *check );
            ++check;
        }
    } );
}

TEST_CASE( "iterators" ) {
    test_list< int > ll;
    test_list< int >::iterator i;
    test_list< int >::const_iterator ci;
    i = ll.begin();
    REQUIRE( i == ll.begin() );
    REQUIRE( i == ll.end() );
    ci = ll.begin();
    REQUIRE( ci == ll.cbegin() );
    REQUIRE( ci == ll.cend() );

    const auto &cll = ll;
    ci = cll.begin();
    REQUIRE( ci == cll.begin() );
    REQUIRE( ci == cll.end() );

    ll.push_back( 1 );
    ll.push_back( 2 );
    i = ll.begin();
    REQUIRE( *++i == 2 );
    REQUIRE( *i++ == 2 );
    REQUIRE( i == ll.end() );

    test_list< int >::const_iterator ci2 = ll.begin();
    REQUIRE( ci2 == ll.cbegin() );
    REQUIRE( ci2 == ll.begin() );
    REQUIRE( ll.begin() == ci2 );
}
