#ifdef EXAMPLE_SOLUTION
#include "ranges_solution.h"
#else
#include "ranges.h"
#endif
#include <catch2/catch.hpp>
#include <initializer_list>
#include <vector>
#include <string>
#include <string_view>
#include <iostream>

using namespace std::literals;
using StringVec = std::vector< std::string >;

template< typename R >
auto materialise( R range ) {
    // this is rather nasty to take into account the following
    // * we want the result to be std::string if value type of rage is char
    // * and std::vector otherwise
    // * there seems to be a bug in the standard library where some results of
    //   std::views::take do not satisfy std::ranges::range
    using value_t = std::remove_cvref_t< decltype( *begin( range ) ) >;
    using out_t = std::conditional_t<
            std::same_as< value_t, char >,
                std::string,
                std::vector< value_t > >;
    out_t out;
    if constexpr ( std::ranges::sized_range< R > ) {
        out.reserve( std::ranges::size( range ) );
    }
    for ( const auto &v : range )
        out.push_back( v );
    return out;
}


TEST_CASE( "getLonger" ) {
    REQUIRE( materialise( getLonger( StringVec{ "hi", "ahoj", "hello"}, 3 ) )
             == StringVec{ "ahoj", "hello" } );
    REQUIRE( materialise( getLonger( std::vector{ std::vector{ 1, 2, 3 }, { 1 }, { } }, 2 ) )
             == std::vector({ std::vector{ 1, 2, 3 } }) );
}

TEST_CASE( "summarizePassing" ) {
    REQUIRE( materialise( summarizePassing( std::vector{ std::tuple{ "A", 42 }, { "B", 1 }, { "C", 25 } }, 25 ) )
             == StringVec{ "A: 42 pts", "C: 25 pts" } );
}

TEST_CASE( "everyNth" ) {
    REQUIRE( materialise( everyNth( "hello"sv, 2 ) ) == "hlo" );
    REQUIRE( materialise( everyNth( "abcdefgh"sv, 3 ) ) == "adg" );
    REQUIRE( materialise( everyNth( std::vector{ 0, 1, 2, 3, 4, 5, 6 }, 3 ) )
             == std::vector{ 0, 3, 6 } );
}

TEST_CASE( "diagSeq" ) {
    REQUIRE( materialise( diagSeq() | std::views::take( 11 ) )
             == std::vector< size_t >{ 0, 0, 1, 0, 1, 2, 0, 1, 2, 3, 0 } );
    REQUIRE( materialise( diagSeq< int >() | std::views::take( 11 ) )
             == std::vector< int >{ 0, 0, 1, 0, 1, 2, 0, 1, 2, 3, 0 } );
}

TEST_CASE( "diagSums" ) {
    REQUIRE( materialise( diagSums() | std::views::take( 11 ) )
             == std::vector< size_t >{ 0, 0, 1, 1, 2, 4, 4, 5, 7, 10, 10 } );
    REQUIRE( materialise( diagSums< int >() | std::views::take( 11 ) )
             == std::vector< int >{ 0, 0, 1, 1, 2, 4, 4, 5, 7, 10, 10 } );
}
